/* -*- mode: C++; c-file-style: "gnu"; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
//#include "base.hh"

#include <string>
#include <stdexcept>

#include <CoreServices/CoreServices.h>
#include <CoreFoundation/CoreFoundation.h>

#include <boost/scoped_array.hpp>

#ifndef __APPLE_CC__

void
platform_fs_normalisation_hook( std::string & const in, std::string out )
{
  out.assign( in );
}

#else

static void
named_directory_fsref( FSRef & out_fsref, char const * const in_name ) {
  OSStatus result = FSPathMakeRef( (UInt8*) in_name, &out_fsref, NULL );
  if( result ) 
    {
      throw std::runtime_error( "failed to make a FSRef for the named directory (named_directory_fsref)" );
    }
}

/* because string.find is dumb */
static std::string::size_type
my_find( std::string const & in_string, char const in_match, std::string::size_type in_start_index ) 
{
  std::string::size_type i = in_string.find( in_match, in_start_index );
  if( i == std::string::npos ) 
    {
      i = in_string.length();
    }
  
  return i;
}

static void
fsref_with_parent_and_name( FSRef const & in_parent, std::string const & in_name, FSRef & out_fsref )
{
  CFStringRef cf_string = CFStringCreateWithFileSystemRepresentation( NULL, in_name.c_str() );
  if( !cf_string ) 
    {
      throw std::runtime_error( "failed to create a CFString with the filename c-string (fsref_with_parent_and_name)" );
    }

  UniCharCount length = CFStringGetLength( cf_string );
  boost::scoped_array<UniChar> buffer( new UniChar[ length ] );
  CFRange range;
  range.location = 0;
  range.length = length;
  CFStringGetCharacters( cf_string, range, buffer.get() );
  CFRelease( cf_string );
  
  OSErr err = FSMakeFSRefUnicode( &in_parent, length, buffer.get(), kTextEncodingUnknown, &out_fsref );
  if( err ) 
    {
      throw std::runtime_error( "failed to make a FSRef using the supplied name (fsref_with_parent_and_name)" );
    }
}

static void
name_for_fsref( FSRef const & in_fsref, std::string& out_name )
{
  HFSUniStr255 actual_filename;
  OSErr err = FSGetCatalogInfo( &in_fsref, kFSCatInfoNone, NULL, &actual_filename, NULL, NULL );
  if( err ) 
    {
      throw std::runtime_error( "failed to fetch the HFSUniStr255 with FSGetCatalogInfo (name_for_fsref)" );
    }
  
  CFStringRef cf_string = CFStringCreateWithCharacters( NULL, actual_filename.unicode, actual_filename.length );
  if( !cf_string ) 
    {
      throw std::runtime_error( "failed to make a CFString from the HFSUniStr255 (name_for_fsref)" );
    }
  
  CFIndex max_representation_size = CFStringGetMaximumSizeOfFileSystemRepresentation( cf_string );
  boost::scoped_array<char> buffer( new char[ max_representation_size ] );
  Boolean result = CFStringGetFileSystemRepresentation( cf_string, buffer.get(), max_representation_size );
  CFRelease( cf_string );
  if( !result ) 
    {
      throw std::runtime_error( "failed to fetch c string from the CFString (name_for_fsref)" );
    }
  
  out_name.assign( buffer.get() );
}

static void
do_normalise( std::string const & in_path, std::string & out_normalised )
{
  std::string::size_type length = in_path.length(), index, previous_index;
  if( length == 0 ) 
    {
      out_normalised.clear();
      return;
    }
  
  FSRef parent_fsref;
  if( in_path[ 0 ] == '/' )
    {
      named_directory_fsref( parent_fsref, "/" );
      index = previous_index = 1;
      out_normalised.assign( "/" );
    }
  else
    {
      named_directory_fsref( parent_fsref, "." );
      index = previous_index = 0;
      out_normalised.clear();
    }

  while( previous_index < length ) 
    {
      index = my_find( in_path, '/', previous_index );
      std::string path_component( in_path, previous_index, index - previous_index );
      FSRef temp_fsref;
      fsref_with_parent_and_name( parent_fsref, path_component, temp_fsref );
      parent_fsref = temp_fsref;
      name_for_fsref( parent_fsref, path_component );
      out_normalised.append( path_component );
      if( index < length ) {
        out_normalised.append( "/" );
      }
      
      previous_index = index + 1;
    }
}

void
platform_fs_normalisation_hook( std::string const & in_path, std::string & out_normalised )
{
  try 
    {
      do_normalise( in_path, out_normalised );
    }
  catch( std::exception e ) 
    {
      /* swallow exceptions so that errors don't cause problems with surrounding code */
      //L(FL("warning: exception in platform_fs_normalisation_hook\n\t(%s)" ) % e.what() ), 
      printf( "warning: exception in platform_fs_normalisation_hook\n\t(%s)\n", e.what() );
      out_normalised.assign( in_path );
    }
  catch( ... ) 
    {
      //L(FL("warning: unknown exception in platform_fs_normalisation_hook ) ), 
      printf( "warning: unknown exception in platform_normalisation_hook\n" );
      out_normalised.assign( in_path );
    }
}


#endif
